import cv2

from color_tracker.utils.camera.base_camera import Camera


class WebCamera(Camera):
    """
    Simple Webcamera
    """

    def __init__(self, video_src=0):
        """
        :param video_src (int): camera source code. It can be an integer or the name of the video file.
        """

        super().__init__()
        self._video_src = video_src

    def _init_camera(self):
        super()._init_camera()
        self._cam = cv2.VideoCapture(self._video_src)
        self._ret, self._frame = self._cam.read()
        if not self._ret:
            raise Exception("No camera feed")
        self._frame_height, self._frame_width, c = self._frame.shape
        return self._ret

    def _read_from_camera(self):
        super()._read_from_camera()
        self._ret, self._frame = self._cam.read()
        if self._ret:
            if self._auto_undistortion:
                self._frame = self._undistort_image(self._frame)
            return True, self._frame
        else:
            return False, None

    def release(self):
        super().release()
        self._cam.release()



import math
from collections import deque
from types import FunctionType

import cv2

from color_tracker.utils import helpers

_RGB_TYPE = "rgb"
_BGR_TYPE = "bgr"
_GRAY_TYPE = "gray"

_ACCEPTED_IMAGE_TYPES = [_RGB_TYPE, _BGR_TYPE, _GRAY_TYPE]


class ColorTracker(object):
    def __init__(self, camera, max_nb_of_points=None, debug=True):
        """
        :param camera: Camera object which parent is a Camera object (like WebCamera)
        :param max_nb_of_points: Maxmimum number of points for storing. If it is set
        to None than it means there is no limit
        :param debug: When it's true than we can see the visualization of the captured points etc...
        """

        super().__init__()
        self._camera = camera
        self._tracker_points = None
        self._debug = debug
        self._max_nb_of_points = max_nb_of_points
        self._selection_points = None
        self._tracking_callback = None
        self._last_detected_object_contour = None
        self._last_detected_object_center = None
        self._object_bounding_box = None
        self._is_running = False
        self._frame = None
        self._debug_frame = None

        self._frame_preprocessor = None

        self._create_tracker_points_list()

    def set_frame_preprocessor(self, preprocessor_func):
        self._frame_preprocessor = preprocessor_func

    def set_court_points(self, court_points):
        """
        Set a set of points that crops out a convex polygon from the image.
        So only on the cropped part will be detection
        :param court_points (list): list of points
        """

        self._selection_points = court_points

    def set_tracking_callback(self, tracking_callback):
        if not isinstance(tracking_callback, FunctionType):
            raise Exception("tracking_callback is not a valid Function with type: FunctionType!")
        self._tracking_callback = tracking_callback

    def _create_tracker_points_list(self):
        """
        Initialize the tracker point list
        """

        if self._max_nb_of_points:
            self._tracker_points = deque(maxlen=self._max_nb_of_points)
        else:
            self._tracker_points = deque()

    def get_tracker_points(self):
        """
        :return (list): Returns the tracker points what were captured
        """
        return self._tracker_points

    def _add_new_tracker_point(self, min_point_distance, max_point_distance):
        try:
            dst = helpers.calculate_distance(self._tracker_points[-1], self._last_detected_object_center)
            if max_point_distance > dst > min_point_distance:
                self._tracker_points.append(self._last_detected_object_center)
        except IndexError:
            # It happens only when the queue is empty and we need a starting point
            self._tracker_points.append(self._last_detected_object_center)

    def _find_and_track_object_center_point(self, contours, min_contour_area,
                                            min_point_distance, max_point_distance):

        self._last_detected_object_contour = helpers.get_largest_contour(contours, min_contour_area)

        if self._last_detected_object_contour is not None:
            self._last_detected_object_center = helpers.get_contour_center(self._last_detected_object_contour)
            self._object_bounding_box = helpers.get_bounding_box_for_contour(self._last_detected_object_contour)
            self._add_new_tracker_point(min_point_distance, max_point_distance)
        else:
            self._last_detected_object_center = None

    def _draw_debug_things(self, draw_tracker_points=True, draw_contour=True,
                           draw_object_center=True, draw_boundin_box=True):
        if draw_contour:
            if self._last_detected_object_contour is not None:
                cv2.drawContours(self._debug_frame, [self._last_detected_object_contour], -1, (0, 255, 0), cv2.FILLED)
        if draw_object_center:
            if self._last_detected_object_center is not None:
                cv2.circle(self._debug_frame, self._last_detected_object_center, 3, (0, 0, 255), -1)
        if draw_boundin_box:
            if self._object_bounding_box is not None:
                pt1 = self._object_bounding_box[0]
                pt2 = self._object_bounding_box[1]
                cv2.rectangle(self._debug_frame, pt1, pt2, (255, 255, 255), 1)
        if draw_tracker_points:
            self._draw_tracker_points(self._debug_frame)

    def clear_track_points(self):
        """
        Delete all tracker points
        """

        if len(self._tracker_points) > 0:
            self._create_tracker_points_list()

    def _draw_tracker_points(self, debug_image):
        if debug_image is not None:
            for i in range(1, len(self._tracker_points)):
                if self._tracker_points[i - 1] is None or self._tracker_points[i] is None:
                    continue
                rectangle_offset = 4
                rectangle_pt1 = tuple(x - rectangle_offset for x in self._tracker_points[i])
                rectangle_pt2 = tuple(x + rectangle_offset for x in self._tracker_points[i])
                cv2.rectangle(debug_image, rectangle_pt1, rectangle_pt2, (255, 255, 255), 1)
                cv2.line(debug_image, self._tracker_points[i - 1], self._tracker_points[i], (255, 255, 255), 1)

    def stop_tracking(self):
        """
        Stop the color tracking
        """

        self._is_running = False

    def _read_from_camera(self, flip_camera_image):
        ret, self._frame = self._camera.read()

        if ret:
            if flip_camera_image:
                self._frame = cv2.flip(self._frame, 1)
        else:
            import warnings
            warnings.warn("There is no camera feed!")

    def track(self, hsv_lower_value, hsv_upper_value, min_contour_area, input_image_type="bgr", kernel=None,
              min_track_point_distance=20, flip_camera_image=True):
        """
        With this we can start the tracking with the given parameters
        :param input_image_type: Type of the input image (color ordering). The standard is BGR because of OpenCV.
        That is the default image ordering but if you use a different type you have to set it here.
        (For example when you use a different input source or you used some preprocessing on the input image)
        :param hsv_lower_value: lowest acceptable hsv values
        :param hsv_upper_value: highest acceptable hsv values
        :param min_contour_area: minimum contour area for the detection. Below that the detection does not count
        :param kernel: structuring element to perform morphological operations on the mask image
        :param min_track_point_distance: minimum distance between the tracked and recognized points
        """

        self._is_running = True

        while True:
            self._read_from_camera(flip_camera_image)

            if self._frame_preprocessor is not None:
                self._frame = self._frame_preprocessor(self._frame)

            self._check_and_fix_image_type(input_image_type=input_image_type)

            if (self._selection_points is not None) and (self._selection_points != []):
                self._frame = helpers.crop_out_polygon_convex(self._frame, self._selection_points)

            img = self._frame.copy()
            self._debug_frame = self._frame.copy()

            cnts = helpers.find_object_contours(image=img,
                                                hsv_lower_value=hsv_lower_value,
                                                hsv_upper_value=hsv_upper_value,
                                                kernel=kernel)

            self._find_and_track_object_center_point(contours=cnts,
                                                     min_contour_area=min_contour_area,
                                                     min_point_distance=min_track_point_distance,
                                                     max_point_distance=math.inf)

            if self._debug:
                self._draw_debug_things(draw_contour=False)

            if self._tracking_callback is not None:
                try:
                    self._tracking_callback()
                except TypeError:
                    import warnings
                    warnings.warn(
                        "Tracker callback function is not working because of wrong arguments! It takes zero arguments")

            if not self._is_running:
                break

    def _check_and_fix_image_type(self, input_image_type="bgr"):
        input_image_type = input_image_type.lower()

        if input_image_type not in _ACCEPTED_IMAGE_TYPES:
            raise ValueError(
                "Image type: {0} is not in accepted types: {1}".format(input_image_type, _ACCEPTED_IMAGE_TYPES))

        try:
            if input_image_type == "rgb":
                self._frame = cv2.cvtColor(self._frame, cv2.COLOR_RGB2BGR)
            elif input_image_type == "gray":
                self._frame = cv2.cvtColor(self._frame, cv2.COLOR_GRAY2BGR)
        except cv2.error as e:
            print("Could not convert to BGR image format. Maybe you should define another input_image_type")
            raise

    def get_debug_image(self):
        if self._debug:
            return self._debug_frame
        else:
            import warnings
            warnings.warn("Debugging is not enabled so there is no debug frame")

    def get_frame(self):
        return self._frame

    def get_last_object_center(self):
        return self._last_detected_object_center

    def get_object_bounding_box(self):
        return self._object_bounding_box

    def get_object_sub_image(self):
        """
        Returns the sub image from the original image defined by the object's bounding box
        """

        pt1 = self._object_bounding_box[0]
        pt2 = self._object_bounding_box[1]
        sub_image = self._frame[pt1[1]:pt2[1], pt1[0]:pt2[0]]
        if len(sub_image) == 0 or sub_image == []:
            return None
        return sub_image



import cv2
import numpy as np


def calculate_distance(pt1, pt2):
    """
    Calculate the distance between two points
    :param pt1: tuple , 2D point
    :param pt2:  tuple, 2D point
    :return: distance between given points
    """

    x1, y1 = pt1
    x2, y2 = pt2
    return np.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)


def crop_out_polygon_convex(image, point_array):
    """
    Crops out a convex polygon given from a list of points from an image
    :param image: Opencv BGR image
    :param point_array: list of points that defines a convex polygon
    :return: Cropped out image
    """

    point_array = np.array(point_array)
    point_array = np.reshape(cv2.convexHull(point_array), point_array.shape)
    mask = np.zeros(image.shape, dtype=np.uint8)
    roi_corners = np.array([point_array], dtype=np.int32)
    channel_count = get_image_channel(image)
    ignore_mask_color = (255,) * channel_count
    cv2.fillConvexPoly(mask, roi_corners, ignore_mask_color)
    masked_image = cv2.bitwise_and(image, mask)
    return masked_image


def get_image_channel(image):
    if len(image.shape) == 3:
        h, w, c = image.shape
    else:
        raise Exception("Image needs to have a shape like: (height, width, channels)")
    return c


def resize_img(image, min_width, min_height):
    """
    Resize the image with keeping the aspect ratio.
    min_width and min_height defines a bounding box
    :param image:
    :param min_width:
    :param min_height:
    :return: resized image
    """
    if len(image.shape) == 3:
        h, w, c = image.shape
    elif len(image.shape) == 2:
        h, w = image.shape
        c = 1
    else:
        raise Exception("Something is wrong with the image dimensions")

    new_w = w
    new_h = h

    if w > min_width:
        new_w = min_width
        new_h = int(h * (float(new_w) / w))

    h, w, c = (new_h, new_w, c)
    if h > min_height:
        new_h = min_height
        new_w = int(w * (float(new_h) / h))

    resized = cv2.resize(image, (new_w, new_h))

    return resized


def get_largest_contour(contours, min_contour_area):
    """
    Find the largest contour in a set of detected contours
    :param contours: detected contours
    :param min_contour_area: only above this threshold it counts as a valid contour
    :return: contour
    """

    if len(contours) > 0:
        c = max(contours, key=cv2.contourArea)
        area = cv2.contourArea(c)
        if area >= min_contour_area:
            return c
    return None


def get_contour_center(contour):
    """
    Calculate the center of the contour
    :param contour: Contour detected with find_contours
    :return: object center as tuple
    """

    # ((x, y), radius) = cv2.minEnclosingCircle(c)
    M = cv2.moments(contour)
    center = (int(M["m10"] / M["m00"]), int(M["m01"] / M["m00"]))
    return center


def find_object_contours(image, hsv_lower_value, hsv_upper_value, kernel):
    hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
    mask = cv2.inRange(hsv, hsv_lower_value, hsv_upper_value)
    if kernel is not None:
        mask = cv2.morphologyEx(mask, cv2.MORPH_CLOSE, kernel, iterations=1)
    return cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)[-2]


def get_bounding_box_for_contour(contour):
    x, y, w, h = cv2.boundingRect(contour)
    pt_top_left = (x, y)
    pt_bottom_right = (x + w, y + h)
    return pt_top_left, pt_bottom_right



import cv2
import imageio

from color_tracker.utils.camera.base_camera import Camera
from color_tracker.utils.helpers import resize_img


class VideoReader(Camera):
    def __init__(self, video_name):
        super().__init__()
        self._current_frame_number = 0
        self._full_length = 0
        self._video_name = video_name

    def _init_camera(self):
        super()._init_camera()
        self._cam = imageio.get_reader(self._video_name, "ffmpeg")
        self._full_length = self._cam.get_length()
        try:
            self._frame = self._cam.get_data(self._current_frame_number)
            self._frame = self._convert_frame_from_rgb_to_bgr(self._cam.get_data(self._current_frame_number))
            self._ret = True
        except Exception:
            self._ret = False
        if not self._ret:
            raise Exception("No camera feed")
        self._frame_height, self._frame_width, c = self._frame.shape
        return self._ret

    def _read_from_camera(self):
        super()._read_from_camera()
        try:
            self._frame = resize_img(
                cv2.flip(self._convert_frame_from_rgb_to_bgr(self._cam.get_data(self._current_frame_number)),
                         1), 640, 480)
            self._ret = True
            self._current_frame_number += 1
            if self._full_length - 50 <= self._current_frame_number:
                self._current_frame_number = 0
        except Exception:
            self._ret = False

        if self._ret:
            if self._auto_undistortion:
                self._frame = self._undistort_image(self._frame)
            return True, self._frame
        else:
            return False, None

    def _convert_frame_from_rgb_to_bgr(self, image):
        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        return image

    def release(self):
        super().release()



import threading

import cv2


class Camera(object):
    """
    Base Camera object
    """

    def __init__(self):
        self._cam = None
        self._frame = None
        self._frame_width = None
        self._frame_height = None
        self._ret = False

        self._auto_undistortion = False
        self._camera_matrix = None
        self._distortion_coefficients = None

        self._is_running = False

    def _init_camera(self):
        """
        This is the first for creating our camera
        We should override this!
        """

        pass

    def start_camera(self):
        """
        Start the running of the camera, without this we can't capture frames
        Camera runs on a separate thread so we can reach a higher FPS
        """

        self._init_camera()
        self._is_running = True
        threading.Thread(target=self._update_camera, args=()).start()

    def _read_from_camera(self):
        """
        This method is responsible for grabbing frames from the camera
        We should override this!
        """

        if self._cam is None:
            raise Exception("Camera is not started!")

    def _update_camera(self):
        """
        Grabs the frames from the camera
        """

        while True:
            if self._is_running:
                self._ret, self._frame = self._read_from_camera()
            else:
                break

    def get_frame_width_and_height(self):
        """
        Returns the width and height of the grabbed images
        :return (int int): width and height
        """

        return self._frame_width, self._frame_height

    def read(self):
        """
        With this you can grab the last frame from the camera
        :return (boolean, np.array): return value and frame
        """
        if self._is_running:
            return self._ret, self._frame
        else:
            import warnings
            warnings.warn("Camera is not started, you should start it with start_camera()")
            return False, None

    def release(self):
        """
        Stop the camera
        """

        self._is_running = False

    def is_running(self):
        return self._is_running

    def set_calibration_matrices(self, camera_matrix, distortion_coefficients):
        self._camera_matrix = camera_matrix
        self._distortion_coefficients = distortion_coefficients

    def activate_auto_undistortion(self):
        self._auto_undistortion = True

    def deactivate_auto_undistortion(self):
        self._auto_undistortion = False

    def _undistort_image(self, image):
        if self._camera_matrix is None or self._distortion_coefficients is None:
            import warnings
            warnings.warn("Undistortion has no effect because <camera_matrix>/<distortion_coefficients> is None!")
            return image

        h, w = image.shape[:2]
        new_camera_matrix, roi = cv2.getOptimalNewCameraMatrix(self._camera_matrix,
                                                               self._distortion_coefficients, (w, h),
                                                               1,
                                                               (w, h))
        undistorted = cv2.undistort(image, self._camera_matrix, self._distortion_coefficients, None,
                                    new_camera_matrix)
        return undistorted

    def __enter__(self):
        self.start_camera()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()



import cv2
import numpy as np
from color_tracker.utils.camera import Camera
from color_tracker.utils import helpers


class HSVColorRangeDetector:
    """
    Just a helper to determine what kind of lower and upper HSV values you need for the tracking
    """

    def __init__(self, camera):
        assert isinstance(camera, Camera), "camera parameter is not a Camera object!"
        assert camera.is_running(), "camera is not running"

        self._camera = camera
        self._trackbars = []
        self._main_window_name = "HSV color range detector"
        cv2.namedWindow(self._main_window_name)
        self._init_trackbars()

    def _init_trackbars(self):
        trackbars_window_name = "hsv settings"
        cv2.namedWindow(trackbars_window_name)

        h_min_trackbar = _Trackbar("H min", trackbars_window_name, 0, 255)
        s_min_trackbar = _Trackbar("S min", trackbars_window_name, 0, 255)
        v_min_trackbar = _Trackbar("V min", trackbars_window_name, 0, 255)

        h_max_trackbar = _Trackbar("H max", trackbars_window_name, 255, 255)
        s_max_trackbar = _Trackbar("S max", trackbars_window_name, 255, 255)

        kernel_x = _Trackbar("kernel x", trackbars_window_name, 0, 255)
        kernel_y = _Trackbar("kernel y", trackbars_window_name, 0, 255)
        v_max_trackbar = _Trackbar("V max", trackbars_window_name, 255, 255)

        self._trackbars = [h_min_trackbar, s_min_trackbar, v_min_trackbar, h_max_trackbar, s_max_trackbar,
                           v_max_trackbar, kernel_x, kernel_y]

    def _get_trackbar_values(self):
        values = []
        for t in self._trackbars:
            value = t.get_value()
            values.append(value)
        return values

    def detect(self):
        while True:
            ret, self.frame = self._camera.read()

            if ret:
                self.frame = cv2.flip(self.frame, 1)
            else:
                continue

            img = self.frame.copy()

            hsv_img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
            values = self._get_trackbar_values()
            h_min, s_min, v_min = values[:3]
            h_max, s_max, v_max = values[3:6]
            kernel_x, kernel_y = values[6:]

            if kernel_y < 1:
                kernel_y = 1
            if kernel_x < 1:
                kernel_x = 1

            thresh = cv2.inRange(hsv_img, (h_min, s_min, v_min), (h_max, s_max, v_max))

            kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (kernel_x, kernel_y))
            thresh = cv2.morphologyEx(thresh, cv2.MORPH_CLOSE, kernel, iterations=1)
            preview = cv2.bitwise_and(img, img, mask=thresh)

            display_width = 360
            display_height = 240

            font_color = (0, 255, 255)
            font_scale = 0.4
            font_org = (5, 10)

            # Original image
            img_display = helpers.resize_img(img, display_width, display_height)
            cv2.putText(img_display, "Original image", font_org, cv2.FONT_HERSHEY_COMPLEX, font_scale, font_color)
            # Thresholded image
            thresh_display = cv2.cvtColor(helpers.resize_img(thresh, display_width, display_height), cv2.COLOR_GRAY2BGR)
            cv2.putText(thresh_display, "Object map", font_org, cv2.FONT_HERSHEY_COMPLEX, font_scale, font_color)
            # Preview of masked objects
            preview_display = helpers.resize_img(preview, display_width, display_height)
            cv2.putText(preview_display, "Object preview", font_org, cv2.FONT_HERSHEY_COMPLEX, font_scale, font_color)
            # HSV image
            hsv_img_display = helpers.resize_img(hsv_img, display_width, display_height)
            cv2.putText(hsv_img_display, "HSV image", font_org, cv2.FONT_HERSHEY_COMPLEX, font_scale, font_color)

            # Combine images
            display_img_1 = np.concatenate((img_display, thresh_display), axis=1)
            display_img_2 = np.concatenate((preview_display, hsv_img_display), axis=1)
            display_img = np.concatenate((display_img_1, display_img_2), axis=0)

            cv2.imshow(self._main_window_name, display_img)
            key = cv2.waitKey(1)
            if key == 27:
                break

        self._camera.release()
        cv2.destroyAllWindows()

        upper_color = np.array([h_max, s_max, v_max])
        lower_color = np.array([h_min, s_min, v_min])

        return lower_color, upper_color, kernel


class _Trackbar(object):
    def __init__(self, name, parent_window_name, init_value=0, max_value=255):
        self.parent_window_name = parent_window_name
        self.name = name
        self.init_value = init_value
        self.max_value = max_value

        cv2.createTrackbar(self.name, self.parent_window_name, self.init_value, self.max_value, lambda x: x)

    def get_value(self):
        value = cv2.getTrackbarPos(self.name, self.parent_window_name)
        return value



"""
*****************************************************
*               Color Tracker
*
*              Gabor Vecsei
* Email:       vecseigabor.x@gmail.com
* Blog:        https://gaborvecsei.wordpress.com/
* LinkedIn:    https://www.linkedin.com/in/gaborvecsei
* Github:      https://github.com/gaborvecsei
*
*****************************************************
"""

from codecs import open
from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='color_tracker',
    version='0.0.3',
    description='Easy to use color tracking package for object tracking based on colors',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/gaborvecsei/Color-Tracker',
    author='Gabor Vecsei',
    author_email='vecseigabor.x@gmail.com',
    license='MIT',
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Education',
        'Topic :: Software Development',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3'],
    keywords='color tracker vecsei gaborvecsei color_tracker',
    packages=find_packages(),
)
